;; Generated package description from cobol-mode.el  -*- no-byte-compile: t -*-
(define-package "cobol-mode" "1.1.0.20240929.185547" "Mode for editing COBOL code" '((cl-lib "0.5")) :commit "533d2e999fc0e555ccc5819365122c3850620a7e" :url "https://elpa.gnu.org/packages/cobol-mode.html" :authors '(("Edward Hart" . "edward.dan.hart@gmail.com")) :maintainer '("Kristofer Hjelmtorp" . "kristofer@hjelmtorp.se") :keywords '("languages"))
